/*
* JBoss, Home of Professional Open Source
* Copyright 2009, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/
package org.jboss.test.remoting.socketfactory;

import java.io.IOException;
import java.net.InetAddress;
import java.net.Socket;
import java.net.UnknownHostException;
import java.util.HashMap;
import java.util.Map;

import javax.management.MBeanServer;
import javax.net.SocketFactory;

import junit.framework.TestCase;

import org.apache.log4j.ConsoleAppender;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.log4j.PatternLayout;
import org.jboss.remoting.Client;
import org.jboss.remoting.InvocationRequest;
import org.jboss.remoting.InvokerLocator;
import org.jboss.remoting.Remoting;
import org.jboss.remoting.ServerInvocationHandler;
import org.jboss.remoting.ServerInvoker;
import org.jboss.remoting.callback.InvokerCallbackHandler;
import org.jboss.remoting.transport.ClientInvoker;
import org.jboss.remoting.transport.Connector;
import org.jboss.remoting.transport.PortUtil;


/**
 * Unit tests for JBREM-1268.
 * 
 * @author <a href="mailto:ron.sigal@jboss.com">Ron Sigal</a>
 * @version $Rev$
 * <p>
 * Copyright Jan 24, 2011
 * </p>
 */
public class ConfigurationMapChangeTestCase extends TestCase
{
   private static Logger log = Logger.getLogger(ConfigurationMapChangeTestCase.class);
   
   private static boolean firstTime = true;
   
   protected String host;
   protected int port;
   protected String locatorURI;
   protected InvokerLocator serverLocator;
   protected Connector connector;
   protected TestInvocationHandler invocationHandler;

   
   public void setUp() throws Exception
   {
      if (firstTime)
      {
         firstTime = false;
         Logger.getLogger("org.jboss.remoting").setLevel(Level.INFO);
         Logger.getLogger("org.jboss.test.remoting").setLevel(Level.INFO);
         String pattern = "[%d{ABSOLUTE}] [%t] %5p (%F:%L) - %m%n";
         PatternLayout layout = new PatternLayout(pattern);
         ConsoleAppender consoleAppender = new ConsoleAppender(layout);
         Logger.getRootLogger().addAppender(consoleAppender);  
      }
   }

   
   public void tearDown()
   {
   }
   
   
   public void testWithoutChangeRemote() throws Throwable
   {
      log.info("entering " + getName());
      
      // Start server.
      setupServer("");
      
      // Create client.
      InvokerLocator clientLocator = new InvokerLocator(locatorURI);
      HashMap clientConfig = new HashMap();
      clientConfig.put(InvokerLocator.FORCE_REMOTE, "true");
      addExtraClientConfig(clientConfig);
      Client client1 = new Client(clientLocator, clientConfig);
      client1.connect();
      log.info("first client is connected");
      ClientInvoker invoker1 = client1.getInvoker();
      Client client2 = new Client(clientLocator, clientConfig);
      client2.connect();
      ClientInvoker invoker2 = client2.getInvoker();
      log.info("second client is connected");
      
      // Test connections.
      assertEquals("abc", client1.invoke("abc"));
      log.info("first connection is good");
      assertEquals("abc", client2.invoke("abc"));
      log.info("second connection is good");
      
      // Verify invokers are identical.
      assertSame(invoker1, invoker2);
      
      client1.disconnect();
      client2.disconnect();
      shutdownServer();
      log.info(getName() + " PASSES");
   }
   

   public void testWithChangeRemote() throws Throwable
   {
      log.info("entering " + getName());
      
      // Start server.
      setupServer("&" + Remoting.SOCKET_FACTORY_CLASS_NAME + "=" + TestSocketFactory.class.getName());
      
      // Create client.
      InvokerLocator clientLocator = new InvokerLocator(locatorURI);
      HashMap clientConfig = new HashMap();
      clientConfig.put(InvokerLocator.FORCE_REMOTE, "true");
      addExtraClientConfig(clientConfig);
      Client client1 = new Client(clientLocator, clientConfig);
      client1.connect();
      log.info("first client is connected");
      ClientInvoker invoker1 = client1.getInvoker();
      Client client2 = new Client(clientLocator, clientConfig);
      client2.connect();
      ClientInvoker invoker2 = client2.getInvoker();
      log.info("second client is connected");
      
      // Test connections.
      assertEquals("abc", client1.invoke("abc"));
      log.info("first connection is good");
      assertEquals("abc", client2.invoke("abc"));
      log.info("second connection is good");
      
      // Verify invokers are identical.
      assertSame(invoker1, invoker2);
      
      client1.disconnect();
      client2.disconnect();
      shutdownServer();
      log.info(getName() + " PASSES");
   }
   
   
   public void testWithoutChangeLocal() throws Throwable
   {
      log.info("entering " + getName());
      
      // Start server.
      setupServer("");
      
      // Create client.
      InvokerLocator clientLocator = new InvokerLocator(locatorURI);
      HashMap clientConfig = new HashMap();
      addExtraClientConfig(clientConfig);
      Client client1 = new Client(clientLocator, clientConfig);
      client1.connect();
      log.info("first client is connected");
      ClientInvoker invoker1 = client1.getInvoker();
      Client client2 = new Client(clientLocator, clientConfig);
      client2.connect();
      ClientInvoker invoker2 = client2.getInvoker();
      log.info("second client is connected");
      
      // Test connections.
      assertEquals("abc", client1.invoke("abc"));
      log.info("first connection is good");
      assertEquals("abc", client2.invoke("abc"));
      log.info("second connection is good");
      
      // Verify invokers are identical.
      assertSame(invoker1, invoker2);
      
      client1.disconnect();
      client2.disconnect();
      shutdownServer();
      log.info(getName() + " PASSES");
   }
   

   public void testWithChangeLocal() throws Throwable
   {
      log.info("entering " + getName());
      
      // Start server.
      setupServer("&" + Remoting.SOCKET_FACTORY_CLASS_NAME + "=" + TestSocketFactory.class.getName());
      
      // Create client.
      InvokerLocator clientLocator = new InvokerLocator(locatorURI);
      HashMap clientConfig = new HashMap();
      addExtraClientConfig(clientConfig);
      Client client1 = new Client(clientLocator, clientConfig);
      client1.connect();
      log.info("first client is connected");
      ClientInvoker invoker1 = client1.getInvoker();
      Client client2 = new Client(clientLocator, clientConfig);
      client2.connect();
      ClientInvoker invoker2 = client2.getInvoker();
      log.info("second client is connected");
      
      // Test connections.
      assertEquals("abc", client1.invoke("abc"));
      log.info("first connection is good");
      assertEquals("abc", client2.invoke("abc"));
      log.info("second connection is good");
      
      // Verify invokers are identical.
      assertSame(invoker1, invoker2);
      
      client1.disconnect();
      client2.disconnect();
      shutdownServer();
      log.info(getName() + " PASSES");
   }
   
   
   protected String getTransport()
   {
      return "socket";
   }
   
   
   protected void addExtraClientConfig(Map config) {}
   protected void addExtraServerConfig(Map config) {}
   

   protected void setupServer(String extraParams) throws Exception
   {
      host = InetAddress.getLocalHost().getHostAddress();
      port = PortUtil.findFreePort(host);
      locatorURI = getTransport() + "://" + host + ":" + port + "/?timeout=10000&x=y" + extraParams;
      String metadata = System.getProperty("remoting.metadata");
      if (metadata != null)
      {
         locatorURI += "&" + metadata;
      }
      serverLocator = new InvokerLocator(locatorURI);
      log.info("Starting remoting server with locator uri of: " + locatorURI);
      HashMap config = new HashMap();
      config.put(InvokerLocator.FORCE_REMOTE, "true");
      addExtraServerConfig(config);
      connector = new Connector(serverLocator, config);
      connector.create();
      invocationHandler = new TestInvocationHandler();
      connector.addInvocationHandler("test", invocationHandler);
      connector.start();
   }
   
   
   protected void shutdownServer() throws Exception
   {
      if (connector != null)
         connector.stop();
   }
   
   
   static class TestInvocationHandler implements ServerInvocationHandler
   {
      public void addListener(InvokerCallbackHandler callbackHandler) {}
      public Object invoke(final InvocationRequest invocation) throws Throwable
      {
         return invocation.getParameter();
      }
      public void removeListener(InvokerCallbackHandler callbackHandler) {}
      public void setMBeanServer(MBeanServer server) {}
      public void setInvoker(ServerInvoker invoker) {}
   }
   
   
   static class TestSocketFactory extends SocketFactory
   {
      SocketFactory sf = SocketFactory.getDefault();
      
      public Socket createSocket(String arg0, int arg1) throws IOException, UnknownHostException
      {
         return sf.createSocket(arg0, arg1);
      }
      public Socket createSocket(InetAddress arg0, int arg1) throws IOException
      {
         return sf.createSocket(arg0, arg1);
      }
      public Socket createSocket(String arg0, int arg1, InetAddress arg2, int arg3) throws IOException, UnknownHostException
      {
         return sf.createSocket(arg0, arg1, arg2, arg3);
      }
      public Socket createSocket(InetAddress arg0, int arg1, InetAddress arg2, int arg3) throws IOException
      {
         return sf.createSocket(arg0, arg1, arg2, arg3);
      }
   }
}