/*
* JBoss, Home of Professional Open Source
* Copyright 2009, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/
package org.jboss.test.remoting.transport.bisocket.dos;

import java.io.IOException;
import java.lang.reflect.Field;
import java.net.InetAddress;
import java.net.Socket;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import javax.management.MBeanServer;

import junit.framework.TestCase;

import org.apache.log4j.ConsoleAppender;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.log4j.PatternLayout;
import org.jboss.remoting.Client;
import org.jboss.remoting.InvocationRequest;
import org.jboss.remoting.InvokerLocator;
import org.jboss.remoting.ServerInvocationHandler;
import org.jboss.remoting.ServerInvoker;
import org.jboss.remoting.callback.Callback;
import org.jboss.remoting.callback.HandleCallbackException;
import org.jboss.remoting.callback.InvokerCallbackHandler;
import org.jboss.remoting.transport.Connector;
import org.jboss.remoting.transport.PortUtil;
import org.jboss.remoting.transport.bisocket.Bisocket;
import org.jboss.remoting.transport.bisocket.BisocketServerInvoker;

import EDU.oswego.cs.dl.util.concurrent.Semaphore;

/**
 * @author <a href="mailto:ron.sigal@jboss.com">Ron Sigal</a>
 * @version $Rev$
 * <p>
 * Copyright Oct 13, 2010
 * </p>
 */
public class DosTestCase extends TestCase
{
   private static final Logger log = Logger.getLogger(DosTestCase.class);
   private static final String CALLBACK_TEST = "callbackTest";
   private static final int dosMaxThreadsValue = 49;
   private static final int dosTimeoutValue = 59;
   
   private static boolean firstTime = true;
   
   protected String host;
   protected int port;
   protected int secondaryPort;
   protected String locatorURI;
   protected InvokerLocator serverLocator;
   protected Connector connector;
   protected TestInvocationHandler invocationHandler;
   protected Object lock = new Object();
   protected boolean dosAttackThreadRan;
   protected boolean secondCallbackRan;

   
   public void setUp() throws Exception
   {
      if (firstTime)
      {
         firstTime = false;
         Logger.getLogger("org.jboss.remoting").setLevel(Level.INFO);
         Logger.getLogger("org.jboss.test.remoting").setLevel(Level.INFO);
         String pattern = "[%d{ABSOLUTE}] [%t] %5p (%F:%L) - %m%n";
         PatternLayout layout = new PatternLayout(pattern);
         ConsoleAppender consoleAppender = new ConsoleAppender(layout);
         Logger.getRootLogger().addAppender(consoleAppender);  
      }
   }

   
   public void tearDown()
   {
   }
   
   public void testDosAttack() throws Throwable
   {
      log.info("entering " + getName());
      
      // Start server.
      setupServer(false, false);
      
      // Create client.
      HashMap clientConfig = new HashMap();
      clientConfig.put(InvokerLocator.FORCE_REMOTE, "true");
      addExtraClientConfig(clientConfig);
      final Client client = new Client(serverLocator, clientConfig);
      client.connect();
      assertEquals("abc", client.invoke("abc"));
      log.info("client is connected");
      
      // Add callback handler.
      TestCallbackHandler callbackHandler = new TestCallbackHandler();
      HashMap metadata = new HashMap();
      metadata.put(Bisocket.IS_CALLBACK_SERVER, "true");
      client.addListener(callbackHandler, metadata);
      client.invoke(CALLBACK_TEST);
      assertEquals(1, callbackHandler.counter);
      log.info("callback handler is installed");
      
      // Test DOS attack. 
      new Thread()
      {
         public void run()
         {
            try
            {
               Socket s = new Socket(host, secondaryPort);
               log.info(this + " created socket " + s);
               synchronized (lock)
               {
                  dosAttackThreadRan = true;
               } 
            }
            catch (IOException e)
            {
               log.error("unable to connect to secondaryPort: " + secondaryPort, e);
            }
            finally
            {
               synchronized (lock)
               {
                  lock.notifyAll();
               }  
            }
         }
      }.start();
      
      Thread.sleep(2000);
      
      synchronized (lock)
      {
         if (!dosAttackThreadRan)
         {
            long start = System.currentTimeMillis();
            long end = start + 10000;
            while (end - System.currentTimeMillis() > 0)
            {
               try
               {
                  lock.wait(end - System.currentTimeMillis());
               }
               catch (InterruptedException e){

               }
            }
         }
      } 
      
      if (!dosAttackThreadRan)
      {
         fail("DOS attack thread did not run");
      }
      
      
      // DOS attack has occurred.  Try to add another callback handler.
      new Thread()
      {
         public void run()
         {
            TestCallbackHandler callbackHandler2 = new TestCallbackHandler();
            try
            {
               
               HashMap metadata = new HashMap();
               metadata.put(Bisocket.IS_CALLBACK_SERVER, "true");
               client.addListener(callbackHandler2, metadata);
               secondCallbackRan = true;
               log.info(this + " second callback handler installed after DOS attack");
            }
            catch (Throwable e)
            {
               log.info(this + " second callback failed", e);
            }    
         }
      }.start();
      
      Thread.sleep(10000);
      assertTrue(secondCallbackRan);
      
      client.removeListener(callbackHandler);
      client.disconnect();
      connector.stop();
   }
   
   
   public void testConfigurationDefault() throws Throwable
   {
      log.info("entering " + getName());
      doConfigurationTest(false, false, Bisocket.SECONDARY_MAX_THREADS_DEFAULT, Bisocket.SECONDARY_TIMEOUT_DEFAULT);
      log.info(getName() + " PASSES");
   }

   
   public void testConfigurationMap() throws Throwable
   {
      log.info("entering " + getName());
      doConfigurationTest(true, false, dosMaxThreadsValue, dosTimeoutValue);
      log.info(getName() + " PASSES");
   }
   
   
   public void testConfigurationInvokerLocater() throws Throwable
   {
      log.info("entering " + getName());
      doConfigurationTest(true, true, dosMaxThreadsValue, dosTimeoutValue);
      log.info(getName() + " PASSES");
   }
   
   
   protected void doConfigurationTest(boolean setParameters, boolean useInvokerLocator, int threadCount, int timeout) throws Throwable
   {
      // Start server.
      setupServer(setParameters, useInvokerLocator);
      
      // Create client.
      HashMap clientConfig = new HashMap();
      clientConfig.put(InvokerLocator.FORCE_REMOTE, "true");
      addExtraClientConfig(clientConfig);
      final Client client = new Client(serverLocator, clientConfig);
      client.connect();
      assertEquals("abc", client.invoke("abc"));
      log.info("client is connected");
      
      // Add callback handler.
      TestCallbackHandler callbackHandler = new TestCallbackHandler();
      HashMap metadata = new HashMap();
      metadata.put(Bisocket.IS_CALLBACK_SERVER, "true");
      client.addListener(callbackHandler, metadata);
      client.invoke(CALLBACK_TEST);
      assertEquals(1, callbackHandler.counter);
      log.info("callback handler is installed");
      
      BisocketServerInvoker invoker = (BisocketServerInvoker) connector.getServerInvoker();
      assertEquals(threadCount, invoker.getDosMaxThreads());
      assertEquals(timeout, invoker.getDosTimeout());
      verifyThreadValues(invoker, threadCount, timeout);

      client.removeListener(callbackHandler);
      client.disconnect();
      connector.stop();
   }
   
   
   protected boolean verifyThreadValues(ServerInvoker invoker, int threadCount, int timeout) throws Exception
   {
      Class[] classes = BisocketServerInvoker.class.getDeclaredClasses();
      Class threadClass = null;
      for (int i = 0; i < classes.length; i++)
      {
         if (classes[i].getName().indexOf("SecondaryServerSocketThread") > -1)
         {
            threadClass = classes[i];
            break;
         }
      }
      log.info("threadClass: " + threadClass);
      Field secondaryServerSocketThreads = BisocketServerInvoker.class.getDeclaredField("secondaryServerSocketThreads");
      secondaryServerSocketThreads.setAccessible(true);
      Thread secondaryServerSocketThread = (Thread) ((Set)secondaryServerSocketThreads.get(invoker)).iterator().next();
      Field maxThreads = threadClass.getDeclaredField("maxThreads");
      maxThreads.setAccessible(true);
      Field localDosTimeout = threadClass.getDeclaredField("localDosTimeout");
      localDosTimeout.setAccessible(true);
     
      assertEquals(threadCount, ((Semaphore)maxThreads.get(secondaryServerSocketThread)).permits());
      assertEquals(timeout, ((Integer)localDosTimeout.get(secondaryServerSocketThread)).intValue());
      return true;
   }
   
   
   protected String getTransport()
   {
      return "bisocket";
   }
   
   
   protected void addExtraClientConfig(Map config) {}
   protected void addExtraServerConfig(Map config) {}
   

   protected void setupServer(boolean setParameters, boolean useInvokerLocator) throws Exception
   {
      host = InetAddress.getLocalHost().getHostAddress();
      port = PortUtil.findFreePort(host);
      secondaryPort =  PortUtil.findFreePort(host);
      locatorURI = getTransport() + "://" + host + ":" + port + "/?secondaryBindPort=" + secondaryPort;
      String metadata = System.getProperty("remoting.metadata");
      if (metadata != null)
      {
         locatorURI += "&" + metadata;
      }
      HashMap config = new HashMap();
      config.put(InvokerLocator.FORCE_REMOTE, "true");
      addExtraServerConfig(config);
      if (setParameters)
      {
         if (useInvokerLocator)
         {
            locatorURI += "&" + Bisocket.SECONDARY_MAX_THREADS + "=" + Integer.toString(dosMaxThreadsValue);
            locatorURI += "&" + Bisocket.SECONDARY_TIMEOUT + "=" + Integer.toString(dosTimeoutValue);
         }
         else
         {
            config.put(Bisocket.SECONDARY_MAX_THREADS, Integer.toString(dosMaxThreadsValue));
            config.put(Bisocket.SECONDARY_TIMEOUT, Integer.toString(dosTimeoutValue));
         }
      }
      serverLocator = new InvokerLocator(locatorURI);
      log.info("Starting remoting server with locator uri of: " + locatorURI);
      connector = new Connector(serverLocator, config);
      connector.create();
      invocationHandler = new TestInvocationHandler();
      connector.addInvocationHandler("test", invocationHandler);
      connector.start();
   }
   
   
   protected void shutdownServer() throws Exception
   {
      if (connector != null)
         connector.stop();
   }
   
   

   static class TestInvocationHandler implements ServerInvocationHandler
   {
      public Set listeners = new HashSet();
      
      public void addListener(InvokerCallbackHandler callbackHandler)
      {
         listeners.add(callbackHandler);
      }
      public Object invoke(final InvocationRequest invocation) throws Throwable
      {
         if (CALLBACK_TEST.equals(invocation.getParameter()))
         {
            Iterator it = listeners.iterator();
            while (it.hasNext())
            {
               InvokerCallbackHandler handler = (InvokerCallbackHandler) it.next();
               handler.handleCallback(new Callback("test"));
               log.info(this + " sent callback");
            }
         }
         return invocation.getParameter();
      }
      public void removeListener(InvokerCallbackHandler callbackHandler) {}
      public void setMBeanServer(MBeanServer server) {}
      public void setInvoker(ServerInvoker invoker) {}
   }
   
   static class TestCallbackHandler implements InvokerCallbackHandler
   {
      public int counter;
      
      public void handleCallback(Callback callback) throws HandleCallbackException
      {
         log.info(this + " received callback");
         counter++;
      }
   }
}