/*
* JBoss, Home of Professional Open Source
* Copyright 2009, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/
package org.jboss.test.remoting.transport.socket.timeout;

import java.net.InetAddress;
import java.util.HashMap;
import java.util.Map;

import javax.management.MBeanServer;

import junit.framework.TestCase;

import org.apache.log4j.ConsoleAppender;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.log4j.PatternLayout;
import org.jboss.remoting.CannotConnectException;
import org.jboss.remoting.Client;
import org.jboss.remoting.InvocationRequest;
import org.jboss.remoting.InvokerLocator;
import org.jboss.remoting.ServerInvocationHandler;
import org.jboss.remoting.ServerInvoker;
import org.jboss.remoting.callback.InvokerCallbackHandler;
import org.jboss.remoting.transport.Connector;
import org.jboss.remoting.transport.PortUtil;

import EDU.oswego.cs.dl.util.concurrent.CountDown;


/**
 * JBREM-1305
 * 
 * @author <a href="mailto:ron.sigal@jboss.com">Ron Sigal</a>
 * @version $Rev$
 * <p>
 * Copyright Jul 16, 2012
 * </p>
 */
public class SemaphoreGrowthTestCase extends TestCase
{
   private static Logger log = Logger.getLogger(SemaphoreGrowthTestCase.class);
   
   private static boolean firstTime = true;
   
   protected String host;
   protected int port;
   protected String locatorURI;
   protected InvokerLocator serverLocator;
   protected Connector connector;
   protected TestInvocationHandler invocationHandler;
   protected CountDown latch0 = new CountDown(1);
   protected CountDown latch1 = new CountDown(2);
   
   public void setUp() throws Exception
   {
      if (firstTime)
      {
         firstTime = false;
         Logger.getLogger("org.jboss.remoting").setLevel(Level.INFO);
         Logger.getLogger("org.jboss.test.remoting").setLevel(Level.INFO);
         String pattern = "[%d{ABSOLUTE}] [%t] %5p (%F:%L) - %m%n";
         PatternLayout layout = new PatternLayout(pattern);
         ConsoleAppender consoleAppender = new ConsoleAppender(layout);
         Logger.getRootLogger().addAppender(consoleAppender);  
      }
   }

   
   public void tearDown()
   {
   }
   
   
   public void testSemaphoreGrowth() throws Throwable
   {
      log.info("entering " + getName());
      
      // Start server.
      setupServer();
      
      // Create client.
      InvokerLocator clientLocator = new InvokerLocator(locatorURI);
      HashMap clientConfig = new HashMap();
      clientConfig.put(InvokerLocator.FORCE_REMOTE, "true");
      addExtraClientConfig(clientConfig);
      Client client = new Client(clientLocator, clientConfig);
      client.connect();
      log.info("client is connected");
      
      // Make first two calls, attempt to increase semaphore count.
      CallThread ct0 = new CallThread(client, "20000", "longwait", latch0, 0);
      ct0.start();
      Thread.sleep(500);
      try
      {
         log.info("Making second invocation");
         HashMap metadata = new HashMap();
         metadata.put("timeout", "1000");
         client.invoke("unused", metadata);
         fail("Second call unexpectedly succeeded");
      }
      catch (CannotConnectException e)
      {
         if (e.getCause() instanceof IllegalStateException && e.getMessage().indexOf("are in use") > -1)
         {
            log.info("Second call gets expected timeout");
         }
         else
         {
            fail("Second call got unexpected exception wrapped in CannotConnectException: " + e);
         }
      }
      catch (Throwable t)
      {
         fail("Second call got unexpected exception: " + t);
      }
      
      // Verify that semaphore count has not been increased.
      latch0.acquire();
      CallThread ct1 = new CallThread(client, "10000", "shortwait", latch1, 1);
      CallThread ct2 = new CallThread(client, "1000", "nowait", latch1, 2);
      ct1.start();
      Thread.sleep(500);
      ct2.start();
      latch1.acquire();
      log.info("Exception for fourth call: " + ct2.t);
      assertTrue("Fourth call should have timed out",
            ct2.t != null &&
            ct2.t.getCause() instanceof IllegalStateException && 
            ct2.t.getMessage().indexOf("are in use") > -1);
      
      client.disconnect();
      shutdownServer();
      log.info(getName() + " PASSES");
   }
   
   
   protected String getTransport()
   {
      return "socket";
   }
   
   
   protected void addExtraClientConfig(Map config) {}
   protected void addExtraServerConfig(Map config) {}
   

   protected void setupServer() throws Exception
   {
      host = InetAddress.getLocalHost().getHostAddress();
      port = PortUtil.findFreePort(host);
      locatorURI = getTransport() + "://" + host + ":" + port;
      locatorURI += "/?timeout=6000&clientMaxPoolSize=1&numberOfCallRetries=1";
      String metadata = System.getProperty("remoting.metadata");
      if (metadata != null)
      {
         locatorURI += "&" + metadata;
      }
      serverLocator = new InvokerLocator(locatorURI);
      log.info("Starting remoting server with locator uri of: " + locatorURI);
      HashMap config = new HashMap();
      config.put(InvokerLocator.FORCE_REMOTE, "true");
      addExtraServerConfig(config);
      connector = new Connector(serverLocator, config);
      connector.create();
      invocationHandler = new TestInvocationHandler();
      connector.addInvocationHandler("test", invocationHandler);
      connector.start();
   }
   
   
   protected void shutdownServer() throws Exception
   {
      if (connector != null)
         connector.stop();
   }
   
   
   static class TestInvocationHandler implements ServerInvocationHandler
   {
      public void addListener(InvokerCallbackHandler callbackHandler) {}
      public Object invoke(final InvocationRequest invocation) throws Throwable
      {
         Object parameter = invocation.getParameter();
         if ("shortwait".equals(parameter))
         {
            log.info("doing shortwait");
            Thread.sleep(4000);
         }
         else if ("longwait".equals(parameter))
         {
            log.info("doing longwait");
            Thread.sleep(10000);
         }
         return parameter;
      }
      public void removeListener(InvokerCallbackHandler callbackHandler) {}
      public void setMBeanServer(MBeanServer server) {}
      public void setInvoker(ServerInvoker invoker) {}
   }
   
   
   class CallThread extends Thread
   {
      Client client;
      String timeout;
      String command;
      CountDown latch;
      int id;
      Throwable t;
      
      public CallThread(Client client, String timeout, String command, CountDown latch, int id)
      {
         this.client = client;
         this.timeout = timeout;
         this.command = command;
         this.latch = latch;
         this.id = id;
      }
      public void run()
      {
         try
         {
            HashMap metadata = new HashMap();
            metadata.put("timeout", timeout);
            log.info("CallThread " + id + " making invocation: " + command);
            client.invoke(command, metadata);
            log.info("CallThread " + id + " back from invocation");
         }
         catch (Throwable t)
         {
            log.info("Thread " + id + " got exception: " + t.getMessage());
            this.t = t;
         }
         finally
         {
            latch.release();
         }
      }
   }
}